require "defines"

--initialise
function init()
	global.basic_hoppers = {}
	for i = 0,8 do
		global.basic_hoppers[i] = {}
	end
	
	global.fast_hoppers = {}
	for i = 0,4 do
		global.fast_hoppers[i] = {}
	end
	
	global.express_hoppers = {}
	for i = 0,2 do
		global.express_hoppers[i] = {}
	end	
	
	global.smart_hoppers = {}
end
script.on_init(function()
	init()
end)

script.on_configuration_changed(function(config)
	if config.mod_changes ~= nil and config.mod_changes["Logistic-variations"] ~= nil then
		init()
	end
end)


--table keeping
function on_build_basic_hopper(belt)

	local length = #global.basic_hoppers[0] --finds the shortest table.
	local n = 0
	for i = 1,8 do
		if #global.basic_hoppers[i] < length then
			n = i
			length = #global.basic_hoppers[i]
		end
	end
	
	hopper = belt.surface.create_entity{name = "basic-hopper", position = belt.position, force = "player"}
	hopper.operable = false
	table.insert(global.basic_hoppers[n], {hopper = hopper, belt = belt, inventory = hopper.get_inventory(defines.inventory.chest)})
end

function on_build_fast_hopper(belt)

	local length = #global.fast_hoppers[0]
	local n = 0
	for i = 1,4 do
		if #global.fast_hoppers[i] < length then
			n = i
			length = #global.fast_hoppers[i]
		end
	end
	
	hopper = belt.surface.create_entity{name = "fast-hopper", position = belt.position, force = "player"}
	hopper.operable = false
	table.insert(global.fast_hoppers[n], {hopper = hopper, belt = belt, inventory = hopper.get_inventory(defines.inventory.chest)})
end

function on_build_express_hopper(belt)

	local length = #global.express_hoppers[1]
	local n = 0
	for i = 1,2 do
		if #global.express_hoppers[i] < length then
			n = i
			length = #global.express_hoppers[i]
		end
	end
	
	hopper = belt.surface.create_entity{name = "express-hopper", position = belt.position, force = "player"}
	hopper.operable = false
	table.insert(global.express_hoppers[n], {hopper = hopper, belt = belt, inventory = hopper.get_inventory(defines.inventory.chest)})
end

function on_build_smart_hopper(belt)
	hopper = belt.surface.create_entity{name = "smart-hopper", position = belt.position, force = "player"}
	table.insert(global.smart_hoppers, {hopper = hopper, belt = belt, inventory = hopper.get_inventory(defines.inventory.chest)})
end

function on_removed_belt(belt,player)
	local find_entity = belt.surface.find_entity
	local pos = belt.position
	local hopper = find_entity("basic-hopper", pos) or find_entity("fast-hopper", pos) or find_entity("express-hopper", pos) or find_entity("smart-hopper", pos)
	if hopper then 
		if hopper.has_items_inside() then
			for item, count in pairs(hopper.get_inventory(defines.inventory.chest).get_contents()) do
				local removed = 0
				if player then
					removed = player.insert({name = name, count = count})
				end
				if count - removed > 0 then
					hopper.surface.spill_item_stack(hopper.position, {name = name, count = count - removed})
				end
			end
		end
		hopper.destroy() 
	end
end

function on_removed_hopper(hopper)
	local find_entity = hopper.surface.find_entity
	local pos = hopper.position
	local belt = find_entity("basic-hopper-transport-belt", pos) or find_entity("fast-hopper-transport-belt", pos) or find_entity("express-hopper-transport-belt", pos) or find_entity("smart-hopper-transport-belt", pos)
	if belt then belt.destroy() end
end


script.on_event(defines.events.on_built_entity, function(event)
	if event.created_entity.name == "basic-hopper-transport-belt" then 
		on_build_basic_hopper(event.created_entity)
	elseif event.created_entity.name == "fast-hopper-transport-belt" then 
		on_build_fast_hopper(event.created_entity)
	elseif event.created_entity.name == "express-hopper-transport-belt" then 
		on_build_express_hopper(event.created_entity)
	elseif event.created_entity.name == "smart-hopper-transport-belt" then 
		on_build_smart_hopper(event.created_entity)
	end
end)

script.on_event(defines.events.on_robot_built_entity, function(event)
	if event.created_entity.name == "basic-hopper-transport-belt" then 
		on_build_basic_hopper(event.created_entity)
	elseif event.created_entity.name == "fast-hopper-transport-belt" then 
		on_build_fast_hopper(event.created_entity)
	elseif event.created_entity.name == "express-hopper-transport-belt" then 
		on_build_express_hopper(event.created_entity)
	elseif event.created_entity.name == "smart-hopper-transport-belt" then 
		on_build_smart_hopper(event.created_entity)
	end
end)

script.on_event(defines.events.on_entity_died, function(event)
	local name = event.entity.name
	if name == "basic-hopper-transport-belt" or name == "fast-hopper-transport-belt" or name == "express-hopper-transport-belt" or name == "smart-hopper-transport-belt" then 
		on_removed_belt(event.entity)
	end
	if name == "basic-hopper" or name == "fast-hopper" or name == "express-hopper" or "smart-hopper" then 
		on_removed_hopper(event.entity)
	end
end)

script.on_event(defines.events.on_preplayer_mined_item, function(event)
	local name = event.entity.name
	if name == "basic-hopper-transport-belt" or name == "fast-hopper-transport-belt" or name == "express-hopper-transport-belt" or name == "smart-hopper-transport-belt" then 
		on_removed_belt(event.entity)
	end
	if name == "basic-hopper" or name == "fast-hopper" or name == "express-hopper" or "smart-hopper" then 
		on_removed_hopper(event.entity)
	end
end)

script.on_event(defines.events.on_robot_pre_mined, function(event)
	local name = event.entity.name
	if name == "basic-hopper-transport-belt" or name == "fast-hopper-transport-belt" or name == "express-hopper-transport-belt" or name == "smart-hopper-transport-belt" then 
		on_removed_belt(event.entity)
	end
	if name == "basic-hopper" or name == "fast-hopper" or name == "express-hopper" or "smart-hopper" then 
		on_removed_hopper(event.entity)
	end
end)


--item dropping
function drop_on_belt(belt,hopper)
	local r = math.random(2)  	--to 'evenly' spread items on the left and right lane instead of filling one lane first.
								--1 means left line first, 2 means right lane first.
	
	if r == 1 then
		for item,_ in pairs(hopper.get_inventory(defines.inventory.chest).get_contents()) do
			if belt.get_transport_line(defines.transport_line.left_line).insert_at(0.5, {name = item}) then
				hopper.remove_item({name = item})
			end
		end
	end
	
	if r == 2 or hopper.has_items_inside() then	--need to check if the hopper still has item inside, in order to not spawn extra items.
		for item,_ in pairs(hopper.get_inventory(defines.inventory.chest).get_contents()) do
			if belt.get_transport_line(defines.transport_line.right_line).insert_at(0.5, {name = item}) then
				hopper.remove_item({name = item})
			end
		end
	end
	
	if r == 2 and hopper.has_items_inside() then
		for item,_ in pairs(hopper.get_inventory(defines.inventory.chest).get_contents()) do
			if belt.get_transport_line(defines.transport_line.left_line).can_insert_at(0.5) then
				belt.get_transport_line(defines.transport_line.left_line).insert_at(0.5, {name = item})
				hopper.remove_item({name = item})
			end
		end
	end
end

script.on_event(defines.events.on_tick, function(event)

	local n = game.tick % 9	-- every 9th tick = 13.33_ items/sec
	if #global.basic_hoppers[n] > 0 then	--by dividing the hoppers in different tables, the load is shared over multiple ticks
		for i, hopper_data in ipairs(global.basic_hoppers[n]) do
			if not hopper_data.hopper.valid or not hopper_data.belt.valid then
				table.remove(global.basic_hoppers[n],i)
			elseif hopper_data.hopper.has_items_inside() then
				drop_on_belt(hopper_data.belt, hopper_data.hopper)
			end
		end
	end
	
	local n = game.tick % 5 -- every 5th tick = 24 items/sec
	if #global.fast_hoppers[n] > 0 then
		for i, hopper_data in ipairs(global.fast_hoppers[n]) do
			if not hopper_data.hopper.valid or not hopper_data.belt.valid then
				table.remove(global.fast_hoppers[n],i)
			elseif hopper_data.hopper.has_items_inside() then
				drop_on_belt(hopper_data.belt, hopper_data.hopper)
			end
		end
	end
	
	local n = game.tick % 3 -- every 3rd tick = 40 items/sec
	if #global.express_hoppers[n] > 0 then
		for i, hopper_data in ipairs(global.express_hoppers[n]) do
			if not hopper_data.hopper.valid or not hopper_data.belt.valid then
				table.remove(global.express_hoppers[n],i)
			elseif hopper_data.hopper.has_items_inside() then
				drop_on_belt(hopper_data.belt, hopper_data.hopper)
			end
		end
	end
	
	
	if #global.smart_hoppers > 0 then  -- every tick = 120 items/sec
		for i, hopper_data in ipairs(global.smart_hoppers) do
			if not hopper_data.hopper.valid or not hopper_data.belt.valid then
				table.remove(global.smart_hoppers,i)
			elseif hopper_data.hopper.has_items_inside() then
				drop_on_belt(hopper_data.belt, hopper_data.hopper)
			end
		end
	end
	
end)